# encoding=utf-8


"""
Functions for finding the critical value of a particular test statistic given the necessary parameters. Critical
value tables implemented as dictionaries are also provided.

Critical Value Lookup Functions
-------------------------------

.. autosummary::
    :toctree: generated/

    chi_square_critical_value
    d_critical_value
    r_critical_value
    u_critical_value
    w_critical_value

"""

from numpy import nan, array


def chi_square_critical_value(alpha, dof):
    r"""
    Finds the :math:`\chi^2` critical value given the input alpha-level and degrees of freedom.

    Parameters
    ----------
    alpha : float, {0.995, 0.99, 0.975, 0.95, 0.90, 0.10, 0.05, 0.025, 0.01, and 0.005}
        Desired alpha-level for the chi-square critical value.
    dof : int
        Degrees of freedom of the sample observations. Cannot be greater than 30 as critical values are
        not defined.

    Returns
    -------
    chi_crit : float
        The chi-square critical value found from the critical value table.

    Raises
    ------
    ValueError
        If parameter :code:`dof` is greater than 30.
    ValueError
        If parameter :code:`alpha` is not one of (0.995, 0.99, 0.975, 0.95, 0.90, 0.10, 0.05, 0.025, 0.01, and 0.005)

    Examples
    --------
    >>> chi_square_critical_value(0.99, 11)
    3.053
    >>> chi_square_critical_value(0.1, 1)
    2.706

    References
    ----------
    Corder, G.W.; Foreman, D.I. (2014). Nonparametric Statistics: A Step-by-Step Approach.
        Wiley. ISBN 978-1118840313.

    Siegel, S. (1956). Nonparametric statistics: For the behavioral sciences.
        McGraw-Hill. ISBN 07-057348-4

    """
    if isinstance(alpha, str):
        alpha = float(alpha)
    if isinstance(dof, str) or isinstance(dof, int) is False:
        dof = int(dof)

    if dof > 30:
        raise ValueError('chi-square critical value table only provides values for degrees of freedom <= 30.')
    if alpha not in (0.995, 0.99, 0.975, 0.95, 0.90, 0.10, 0.05, 0.025, 0.01, 0.005):
        raise ValueError('chi-square critical value table only provides critical values for alpha levels ' + \
                         '0.995, 0.99, 0.975, 0.95, 0.90, 0.10, 0.05, 0.025, 0.01, and 0.005.')

    chi_crit = _chi_square_critical_value_table[alpha][dof]

    return chi_crit


def d_critical_value(n, alpha):
    if n > 35:
        raise ValueError('critical values of D are only defined for n <= 35.')
    if alpha not in (0.01, 0.05, 0.10, 0.15, 0.20):
        raise ValueError('U critical values are only defined for alpha-levels 0.10, 0.05, 0.025, and 0.01.')

    d_crit = _d_critical_value_table[alpha][n]

    return d_crit


def r_critical_value(n1, n2):
    if not all(array([n1, n2]) <= 20):
        raise ValueError('critical values of the r-statistic are only defined for samples of size 20 or less.')

    r1_crit = _r_critical_value_table['r1'][(n1, n2)]
    r2_crit = _r_critical_value_table['r2'][(n1, n2)]

    return r1_crit, r2_crit


def u_critical_value(n, m, alpha):
    r"""
    Finds the Mann-Whitney :math:`U`-statistic critical value for small sample sizes.

    Parameters
    ----------
    n : int
        Number of observations in first sample group.
    m : int
        Number of observations in second sample group.
    alpha : float, {0.10, 0.05, 0.025, 0.01}
        Alpha-level. Must be one of 0.10, 0.05, 0.025, or 0.01.

    Returns
    -------
    u_crit : int
        The Mann-Whitney U-statistic critical value.

    Raises
    ------
    ValueError
        If parameters :code:`n` or :code:`m` are greater than 20.
    ValueError
        If parameter :code:`alpha` is not one of 0.10, 0.05, 0.025, or 0.01.

    Examples
    --------
    >>> u_critical_value(10, 12, 0.01)
    24
    >>> u_critical_value(8, 8, 0.10)
    19

    References
    ----------
    Corder, G.W.; Foreman, D.I. (2014). Nonparametric Statistics: A Step-by-Step Approach.
        Wiley. ISBN 978-1118840313.

    Siegel, S. (1956). Nonparametric statistics: For the behavioral sciences.
        McGraw-Hill. ISBN 07-057348-4

    """
    if isinstance(alpha, str):
        alpha = float(alpha)
    if isinstance(n, str) or isinstance(n, int) is False:
        n = int(n)
    if isinstance(m, str) or isinstance(m, int) is False:
        m = int(m)

    if n > 20 or m > 20:
        raise ValueError('critical values of U are only defined for n and m <= 20.')
    if alpha not in (0.10, 0.05, 0.025, 0.01):
        raise ValueError('U critical values are only defined for alpha-levels 0.10, 0.05, 0.025, and 0.01.')

    u_crit = _u_critical_value_table[alpha][(n, m)]

    return u_crit


def w_critical_value(n, alpha, alternative):
    r"""
    Finds the :math:`W`-statistic critical value given the input parameters.

    Parameters
    ----------
    n : int
        The number of sample observations. Critical values are only given for :math:`n \geq 30`.
    alpha : float, {'one-tail': {0.05, 0.025, 0.01, 0.005}, 'two-tail': {0.10, 0.05, 0.02, 0.01}}
        Given alpha level.
    alternative : str
        Alternative hypothesis. Must be one of 'one-tail' or 'two-tail'.

    Returns
    -------
    w_crit : float
        The :math:`W`-statistic critical value.

    Raises
    ------
    ValueError
        If parameter :code:`n` is greater than 30.
    ValueError
        If parameter :code:`alpha` is not one of 0.05 or 0.01.
    ValueError
        If parameter :code:`alternative` is not one of 'one-tail' or 'two-tail'.

    Examples
    --------
    >>> w_critical_value(10, 0.05, 'one-tail')
    10
    >>> w_critical_value(11, 0.01, 'two-tail')
    5

    References
    ----------
    Corder, G.W.; Foreman, D.I. (2014). Nonparametric Statistics: A Step-by-Step Approach.
        Wiley. ISBN 978-1118840313.

    Siegel, S. (1956). Nonparametric statistics: For the behavioral sciences.
        McGraw-Hill. ISBN 07-057348-4

    """
    if isinstance(n, str):
        n = int(n)
    if isinstance(alpha, str):
        alpha = float(alpha)

    if n > 30:
        raise ValueError('W critical values are only provided for n <= 30.')
    if alpha not in (0.05, 0.01):
        raise ValueError('W critical values are only provided for alpha values 0.05 and 0.01.')
    if alternative not in ('one-tail', 'two-tail'):
        raise ValueError("alternative must be one of 'one-tail' or 'two-tail'.")

    w_crit = _w_critical_value_table[alternative][alpha][n]

    return w_crit


_chi_square_critical_value_table = {
    0.995: {
        1: nan,
        2: 0.010,
        3: 0.072,
        4: 0.207,
        5: 0.412,
        6: 0.676,
        7: 0.989,
        8: 1.344,
        9: 1.735,
        10: 2.156,
        11: 2.603,
        12: 3.074,
        13: 3.565,
        14: 4.074,
        15: 4.601,
        16: 5.142,
        17: 5.697,
        18: 6.265,
        19: 6.844,
        20: 7.434,
        21: 8.034,
        22: 8.643,
        23: 9.260,
        24: 9.886,
        25: 10.520,
        26: 11.160,
        27: 11.808,
        28: 12.461,
        29: 13.121,
        30: 13.787
    },
    0.99: {
        1: nan,
        2: 0.020,
        3: 0.115,
        4: 0.297,
        5: 0.554,
        6: 0.872,
        7: 1.239,
        8: 1.646,
        9: 2.088,
        10: 2.558,
        11: 3.053,
        12: 3.571,
        13: 4.107,
        14: 4.660,
        15: 5.229,
        16: 5.812,
        17: 6.408,
        18: 7.015,
        19: 7.633,
        20: 8.260,
        21: 8.897,
        22: 9.542,
        23: 10.196,
        24: 10.856,
        25: 11.524,
        26: 12.198,
        27: 12.879,
        28: 13.565,
        29: 14.256,
        30: 14.953
    },
    0.975: {
        1: 0.001,
        2: 0.051,
        3: 0.216,
        4: 0.484,
        5: 0.831,
        6: 1.237,
        7: 1.690,
        8: 2.180,
        9: 2.700,
        10: 3.247,
        11: 3.816,
        12: 4.404,
        13: 5.009,
        14: 5.629,
        15: 6.262,
        16: 6.908,
        17: 7.564,
        18: 8.231,
        19: 8.907,
        20: 9.591,
        21: 10.283,
        22: 10.982,
        23: 11.689,
        24: 12.401,
        25: 13.120,
        26: 13.844,
        27: 14.573,
        28: 15.308,
        29: 16.047,
        30: 16.791
    },
    0.95: {
        1: 0.004,
        2: 0.103,
        3: 0.352,
        4: 0.711,
        5: 1.145,
        6: 1.635,
        7: 2.167,
        8: 2.733,
        9: 3.325,
        10: 3.940,
        11: 4.575,
        12: 5.226,
        13: 5.892,
        14: 6.571,
        15: 7.261,
        16: 7.962,
        17: 8.672,
        18: 9.390,
        19: 10.117,
        20: 10.851,
        21: 11.591,
        22: 12.338,
        23: 13.091,
        24: 13.848,
        25: 14.611,
        26: 15.379,
        27: 16.151,
        28: 16.928,
        29: 17.708,
        30: 18.493
    },
    0.90: {
        1: 0.016,
        2: 0.211,
        3: 0.584,
        4: 1.064,
        5: 1.610,
        6: 2.204,
        7: 2.833,
        8: 3.490,
        9: 4.168,
        10: 4.865,
        11: 5.578,
        12: 6.304,
        13: 7.042,
        14: 7.790,
        15: 8.547,
        16: 9.312,
        17: 10.085,
        18: 10.865,
        19: 11.651,
        20: 12.443,
        21: 13.240,
        22: 14.041,
        23: 14.848,
        24: 15.659,
        25: 16.473,
        26: 17.292,
        27: 18.114,
        28: 18.939,
        29: 19.768,
        30: 20.599
    },
    0.10: {
        1: 2.706,
        2: 4.605,
        3: 6.251,
        4: 7.779,
        5: 9.236,
        6: 10.645,
        7: 12.017,
        8: 13.362,
        9: 14.684,
        10: 15.987,
        11: 17.275,
        12: 18.549,
        13: 19.812,
        14: 21.064,
        15: 22.307,
        16: 23.542,
        17: 24.769,
        18: 25.989,
        19: 27.204,
        20: 28.412,
        21: 29.615,
        22: 30.813,
        23: 32.007,
        24: 33.196,
        25: 34.382,
        26: 35.563,
        27: 36.741,
        28: 37.916,
        29: 39.087,
        30: 40.256
    },
    0.05: {
        1: 3.841,
        2: 5.991,
        3: 7.815,
        4: 9.488,
        5: 11.070,
        6: 12.592,
        7: 14.067,
        8: 15.507,
        9: 16.919,
        10: 18.307,
        11: 19.675,
        12: 21.026,
        13: 22.362,
        14: 23.685,
        15: 24.996,
        16: 26.296,
        17: 27.587,
        18: 28.869,
        19: 30.144,
        20: 31.410,
        21: 32.671,
        22: 33.924,
        23: 35.172,
        24: 36.415,
        25: 37.652,
        26: 38.885,
        27: 40.113,
        28: 41.337,
        29: 42.557,
        30: 43.773
    },
    0.025: {
        1: 5.024,
        2: 7.378,
        3: 9.348,
        4: 11.143,
        5: 12.833,
        6: 14.449,
        7: 16.013,
        8: 17.535,
        9: 19.023,
        10: 20.483,
        11: 21.920,
        12: 23.337,
        13: 24.736,
        14: 26.119,
        15: 27.488,
        16: 28.845,
        17: 30.191,
        18: 31.526,
        19: 32.852,
        20: 34.170,
        21: 35.479,
        22: 36.781,
        23: 38.076,
        24: 39.364,
        25: 40.646,
        26: 41.923,
        27: 43.195,
        28: 44.461,
        29: 45.722,
        30: 46.979
    },
    0.01: {
        1: 6.635,
        2: 9.210,
        3: 11.345,
        4: 13.277,
        5: 15.086,
        6: 16.812,
        7: 18.475,
        8: 20.090,
        9: 21.666,
        10: 23.209,
        11: 24.725,
        12: 26.217,
        13: 27.688,
        14: 29.141,
        15: 30.578,
        16: 32.000,
        17: 33.409,
        18: 34.805,
        19: 36.191,
        20: 37.566,
        21: 38.932,
        22: 40.289,
        23: 21.638,
        24: 42.980,
        25: 44.314,
        26: 45.642,
        27: 46.963,
        28: 48.278,
        29: 49.588,
        30: 50.892
    },
    0.005: {
        1: 7.879,
        2: 10.597,
        3: 12.838,
        4: 14.860,
        5: 16.750,
        6: 18.548,
        7: 20.278,
        8: 21.955,
        9: 23.589,
        10: 25.188,
        11: 26.757,
        12: 28.300,
        13: 29.819,
        14: 31.319,
        15: 32.801,
        16: 34.267,
        17: 35.718,
        18: 37.156,
        19: 38.582,
        20: 39.997,
        21: 41.401,
        22: 42.796,
        23: 44.181,
        24: 45.559,
        25: 46.928,
        26: 48.920,
        27: 49.645,
        28: 50.993,
        29: 52.336,
        30: 53.672
    }
}


_d_critical_value_table = {
    .01: {
        1: .995,
        2: .929,
        3: .828,
        4: .733,
        5: .669,
        6: .618,
        7: .577,
        8: .543,
        9: .514,
        10: .490,
        11: .468,
        12: .450,
        13: .433,
        14: .418,
        15: .404,
        16: .392,
        17: .381,
        18: .371,
        19: .363,
        20: .356,
        25: .320,
        30: .290,
        35: .270
    },
    .05: {
        1: .975,
        2: .842,
        3: .708,
        4: .624,
        5: .565,
        6: .521,
        7: .486,
        8: .457,
        9: .432,
        10: .410,
        11: .391,
        12: .375,
        13: .361,
        14: .349,
        15: .338,
        16: .328,
        17: .318,
        18: .309,
        19: .301,
        20: .294,
        25: .270,
        30: .240,
        35: .230
    },
    .10: {
        1: .950,
        2: .776,
        3: .642,
        4: .564,
        5: .510,
        6: .470,
        7: .438,
        8: .411,
        9: .388,
        10: .368,
        11: .352,
        12: .338,
        13: .325,
        14: .314,
        15: .304,
        16: .295,
        17: .286,
        18: .278,
        19: .272,
        20: .264,
        25: .240,
        30: .220,
        35: .210
    },
    .15: {
        1: .925,
        2: .726,
        3: .597,
        4: .525,
        5: .474,
        6: .436,
        7: .405,
        8: .381,
        9: .360,
        10: .342,
        11: .326,
        12: .313,
        13: .302,
        14: .292,
        15: .283,
        16: .274,
        17: .266,
        18: .259,
        19: .252,
        20: .246,
        25: .220,
        30: .200,
        35: .190
    },
    .20: {
        1: .900,
        2: .684,
        3: .565,
        4: .494,
        5: .446,
        6: .410,
        7: .381,
        8: .358,
        9: .339,
        10: .322,
        11: .307,
        12: .295,
        13: .284,
        14: .274,
        15: .266,
        16: .258,
        17: .250,
        18: .244,
        19: .237,
        20: .231,
        25: .210,
        30: .190,
        35: .180
    }
}


_r_critical_value_table = {
    'r1': {
        (2, 1): nan,
        (2, 2): nan,
        (2, 3): nan,
        (2, 4): nan,
        (2, 5): nan,
        (2, 6): nan,
        (2, 7): nan,
        (2, 8): nan,
        (2, 9): nan,
        (2, 10): nan,
        (2, 11): nan,
        (2, 12): 2,
        (2, 13): 2,
        (2, 14): 2,
        (2, 15): 2,
        (2, 16): 2,
        (2, 17): 2,
        (2, 18): 2,
        (2, 19): 2,
        (2, 20): 2,
        (3, 1): nan,
        (3, 2): nan,
        (3, 3): nan,
        (3, 4): nan,
        (3, 5): nan,
        (3, 6): 2,
        (3, 7): 2,
        (3, 8): 2,
        (3, 9): 2,
        (3, 10): 2,
        (3, 11): 2,
        (3, 12): 2,
        (3, 13): 2,
        (3, 14): 2,
        (3, 15): 3,
        (3, 16): 3,
        (3, 17): 3,
        (3, 18): 3,
        (3, 19): 3,
        (3, 20): 3,
        (4, 1): nan,
        (4, 2): nan,
        (4, 3): nan,
        (4, 4): nan,
        (4, 5): 2,
        (4, 6): 2,
        (4, 7): 2,
        (4, 8): 3,
        (4, 9): 3,
        (4, 10): 3,
        (4, 11): 3,
        (4, 12): 3,
        (4, 13): 3,
        (4, 14): 3,
        (4, 15): 3,
        (4, 16): 4,
        (4, 17): 4,
        (4, 18): 4,
        (4, 19): 4,
        (4, 20): 4,
        (5, 1): nan,
        (5, 2): nan,
        (5, 3): nan,
        (5, 4): 2,
        (5, 5): 2,
        (5, 6): 3,
        (5, 7): 3,
        (5, 8): 3,
        (5, 9): 3,
        (5, 10): 3,
        (5, 11): 4,
        (5, 12): 4,
        (5, 13): 4,
        (5, 14): 4,
        (5, 15): 4,
        (5, 16): 4,
        (5, 17): 4,
        (5, 18): 5,
        (5, 19): 5,
        (5, 20): 5,
        (6, 1): nan,
        (6, 2): nan,
        (6, 3): 2,
        (6, 4): 2,
        (6, 5): 3,
        (6, 6): 3,
        (6, 7): 3,
        (6, 8): 3,
        (6, 9): 4,
        (6, 10): 4,
        (6, 11): 4,
        (6, 12): 4,
        (6, 13): 5,
        (6, 14): 5,
        (6, 15): 5,
        (6, 16): 5,
        (6, 17): 5,
        (6, 18): 5,
        (6, 19): 6,
        (6, 20): 6,
        (7, 1): nan,
        (7, 2): nan,
        (7, 3): 2,
        (7, 4): 2,
        (7, 5): 3,
        (7, 6): 3,
        (7, 7): 3,
        (7, 8): 4,
        (7, 9): 4,
        (7, 10): 5,
        (7, 11): 5,
        (7, 12): 5,
        (7, 13): 5,
        (7, 14): 5,
        (7, 15): 6,
        (7, 16): 6,
        (7, 17): 6,
        (7, 18): 6,
        (7, 19): 6,
        (7, 20): 6,
        (8, 1): nan,
        (8, 2): nan,
        (8, 3): 2,
        (8, 4): 3,
        (8, 5): 3,
        (8, 6): 3,
        (8, 7): 4,
        (8, 8): 4,
        (8, 9): 5,
        (8, 10): 5,
        (8, 11): 5,
        (8, 12): 6,
        (8, 13): 6,
        (8, 14): 6,
        (8, 15): 6,
        (8, 16): 6,
        (8, 17): 7,
        (8, 18): 7,
        (8, 19): 7,
        (8, 20): 7,
        (9, 1): nan,
        (9, 2): nan,
        (9, 3): 2,
        (9, 4): 3,
        (9, 5): 3,
        (9, 6): 4,
        (9, 7): 4,
        (9, 8): 5,
        (9, 9): 5,
        (9, 10): 5,
        (9, 11): 6,
        (9, 12): 6,
        (9, 13): 6,
        (9, 14): 7,
        (9, 15): 7,
        (9, 16): 7,
        (9, 17): 7,
        (9, 18): 8,
        (9, 19): 8,
        (9, 20): 8,
        (10, 1): nan,
        (10, 2): nan,
        (10, 3): 2,
        (10, 4): 3,
        (10, 5): 3,
        (10, 6): 4,
        (10, 7): 5,
        (10, 8): 5,
        (10, 9): 5,
        (10, 10): 6,
        (10, 11): 6,
        (10, 12): 7,
        (10, 13): 7,
        (10, 14): 7,
        (10, 15): 7,
        (10, 16): 8,
        (10, 17): 8,
        (10, 18): 8,
        (10, 19): 8,
        (10, 20): 9,
        (11, 1): nan,
        (11, 2): nan,
        (11, 3): 2,
        (11, 4): 3,
        (11, 5): 4,
        (11, 6): 4,
        (11, 7): 5,
        (11, 8): 5,
        (11, 9): 6,
        (11, 10): 6,
        (11, 11): 7,
        (11, 12): 7,
        (11, 13): 7,
        (11, 14): 8,
        (11, 15): 8,
        (11, 16): 8,
        (11, 17): 9,
        (11, 18): 9,
        (11, 19): 9,
        (11, 20): 9,
        (12, 1): nan,
        (12, 2): 2,
        (12, 3): 2,
        (12, 4): 3,
        (12, 5): 4,
        (12, 6): 4,
        (12, 7): 5,
        (12, 8): 6,
        (12, 9): 6,
        (12, 10): 7,
        (12, 11): 7,
        (12, 12): 7,
        (12, 13): 8,
        (12, 14): 8,
        (12, 15): 8,
        (12, 16): 9,
        (12, 17): 9,
        (12, 18): 9,
        (12, 19): 10,
        (12, 20): 10,
        (13, 1): nan,
        (13, 2): 2,
        (13, 3): 2,
        (13, 4): 3,
        (13, 5): 4,
        (13, 6): 5,
        (13, 7): 5,
        (13, 8): 6,
        (13, 9): 6,
        (13, 10): 7,
        (13, 11): 7,
        (13, 12): 8,
        (13, 13): 8,
        (13, 14): 9,
        (13, 15): 9,
        (13, 16): 9,
        (13, 17): 10,
        (13, 18): 10,
        (13, 19): 10,
        (13, 20): 10,
        (14, 1): nan,
        (14, 2): 2,
        (14, 3): 2,
        (14, 4): 3,
        (14, 5): 4,
        (14, 6): 5,
        (14, 7): 5,
        (14, 8): 6,
        (14, 9): 7,
        (14, 10): 7,
        (14, 11): 8,
        (14, 12): 8,
        (14, 13): 9,
        (14, 14): 9,
        (14, 15): 9,
        (14, 16): 10,
        (14, 17): 10,
        (14, 18): 10,
        (14, 19): 11,
        (14, 20): 11,
        (15, 1): nan,
        (15, 2): 2,
        (15, 3): 3,
        (15, 4): 3,
        (15, 5): 4,
        (15, 6): 5,
        (15, 7): 6,
        (15, 8): 6,
        (15, 9): 7,
        (15, 10): 7,
        (15, 11): 8,
        (15, 12): 8,
        (15, 13): 9,
        (15, 14): 9,
        (15, 15): 10,
        (15, 16): 10,
        (15, 17): 11,
        (15, 18): 11,
        (15, 19): 11,
        (15, 20): 12,
        (16, 1): nan,
        (16, 2): 2,
        (16, 3): 3,
        (16, 4): 4,
        (16, 5): 4,
        (16, 6): 5,
        (16, 7): 6,
        (16, 8): 6,
        (16, 9): 7,
        (16, 10): 8,
        (16, 11): 8,
        (16, 12): 9,
        (16, 13): 9,
        (16, 14): 10,
        (16, 15): 10,
        (16, 16): 11,
        (16, 17): 11,
        (16, 18): 11,
        (16, 19): 12,
        (16, 20): 12,
        (17, 1): nan,
        (17, 2): 2,
        (17, 3): 3,
        (17, 4): 4,
        (17, 5): 4,
        (17, 6): 5,
        (17, 7): 6,
        (17, 8): 7,
        (17, 9): 7,
        (17, 10): 8,
        (17, 11): 9,
        (17, 12): 9,
        (17, 13): 10,
        (17, 14): 10,
        (17, 15): 11, 
        (17, 16): 11,
        (17, 17): 11,
        (17, 18): 12,
        (17, 19): 12,
        (17, 20): 13,
        (18, 1): nan,
        (18, 2): 2,
        (18, 3): 3,
        (18, 4): 4,
        (18, 5): 5,
        (18, 6): 5,
        (18, 7): 6,
        (18, 8): 7,
        (18, 9): 8,
        (18, 10): 8,
        (18, 11): 9,
        (18, 12): 9,
        (18, 13): 10,
        (18, 14): 10,
        (18, 15): 11,
        (18, 16): 11,
        (18, 17): 11,
        (18, 18): 12,
        (18, 19): 12,
        (18, 20): 13,
        (19, 1): nan,
        (19, 2): 2,
        (19, 3): 3,
        (19, 4): 4,
        (19, 5): 5,
        (19, 6): 6,
        (19, 7): 6,
        (19, 8): 7,
        (19, 9): 8,
        (19, 10): 8,
        (19, 11): 9,
        (19, 12): 10,
        (19, 13): 10,
        (19, 14): 11,
        (19, 15): 11,
        (19, 16): 12,
        (19, 17): 12,
        (19, 18): 13,
        (19, 19): 13,
        (19, 20): 13,
        (20, 1): nan,
        (20, 2): 2,
        (20, 3): 3,
        (20, 4): 4,
        (20, 5): 5,
        (20, 6): 6,
        (20, 7): 6,
        (20, 8): 7,
        (20, 9): 8,
        (20, 10): 9,
        (20, 11): 9,
        (20, 12): 10,
        (20, 13): 10,
        (20, 14): 11,
        (20, 15): 12,
        (20, 16): 12,
        (20, 17): 13,
        (20, 18): 13,
        (20, 19): 13,
        (20, 20): 14                 
    },
    'r2': {
        (2, 1): nan,
        (2, 2): nan,
        (2, 3): nan,
        (2, 4): nan,
        (2, 5): nan,
        (2, 6): nan,
        (2, 7): nan,
        (2, 8): nan,
        (2, 9): nan,
        (2, 10): nan,
        (2, 11): nan,
        (2, 12): nan,
        (2, 13): nan,
        (2, 14): nan,
        (2, 15): nan,
        (2, 16): nan,
        (2, 17): nan,
        (2, 18): nan,
        (2, 19): nan,
        (2, 20): nan,
        (3, 1): nan,
        (3, 2): nan,
        (3, 3): nan,
        (3, 4): nan,
        (3, 5): nan,
        (3, 6): nan,
        (3, 7): nan,
        (3, 8): nan,
        (3, 9): nan,
        (3, 10): nan,
        (3, 11): nan,
        (3, 12): nan,
        (3, 13): nan,
        (3, 14): nan,
        (3, 15): nan,
        (3, 16): nan,
        (3, 17): nan,
        (3, 18): nan,
        (3, 19): nan,
        (3, 20): nan,
        (4, 1): nan,
        (4, 2): nan,
        (4, 3): nan,
        (4, 4): nan,
        (4, 5): 9,
        (4, 6): 9,
        (4, 7): nan,
        (4, 8): nan,
        (4, 9): nan,
        (4, 10): nan,
        (4, 11): nan,
        (4, 12): nan,
        (4, 13): nan,
        (4, 14): nan,
        (4, 15): nan,
        (4, 16): nan,
        (4, 17): nan,
        (4, 18): nan,
        (4, 19): nan,
        (4, 20): nan,
        (5, 1): nan,
        (5, 2): nan,
        (5, 3): nan,
        (5, 4): 9,
        (5, 5): 10,
        (5, 6): 10,
        (5, 7): 11,
        (5, 8): 11,
        (5, 9): nan,
        (5, 10): nan,
        (5, 11): nan,
        (5, 12): nan,
        (5, 13): nan,
        (5, 14): nan,
        (5, 15): nan,
        (5, 16): nan,
        (5, 17): nan,
        (5, 18): nan,
        (5, 19): nan,
        (5, 20): nan,
        (6, 1): nan,
        (6, 2): nan,
        (6, 3): nan,
        (6, 4): 9,
        (6, 5): 10,
        (6, 6): 11,
        (6, 7): 12,
        (6, 8): 12,
        (6, 9): 13,
        (6, 10): 13,
        (6, 11): 13,
        (6, 12): 13,
        (6, 13): nan,
        (6, 14): nan,
        (6, 15): nan,
        (6, 16): nan,
        (6, 17): nan,
        (6, 18): nan,
        (6, 19): nan,
        (6, 20): nan,
        (7, 1): nan,
        (7, 2): nan,
        (7, 3): nan,
        (7, 4): nan,
        (7, 5): 11,
        (7, 6): 12,
        (7, 7): 13,
        (7, 8): 13,
        (7, 9): 14,
        (7, 10): 14,
        (7, 11): 14,
        (7, 12): 14,
        (7, 13): 15,
        (7, 14): 15,
        (7, 15): 15,
        (7, 16): nan,
        (7, 17): nan,
        (7, 18): nan,
        (7, 19): nan,
        (7, 20): nan,
        (8, 1): nan,
        (8, 2): nan,
        (8, 3): nan,
        (8, 4): nan,
        (8, 5): 11,
        (8, 6): 12,
        (8, 7): 13,
        (8, 8): 14,
        (8, 9): 14,
        (8, 10): 15,
        (8, 11): 15,
        (8, 12): 16,
        (8, 13): 16,
        (8, 14): 16,
        (8, 15): 16,
        (8, 16): 17,
        (8, 17): 17,
        (8, 18): 17,
        (8, 19): 17,
        (8, 20): 17,
        (9, 1): nan,
        (9, 2): nan,
        (9, 3): nan,
        (9, 4): nan,
        (9, 5): nan,
        (9, 6): 13,
        (9, 7): 14,
        (9, 8): 14,
        (9, 9): 15,
        (9, 10): 16,
        (9, 11): 16,
        (9, 12): 16,
        (9, 13): 17,
        (9, 14): 17,
        (9, 15): 18,
        (9, 16): 18,
        (9, 17): 18,
        (9, 18): 18,
        (9, 19): 18,
        (9, 20): 18,
        (10, 1): nan,
        (10, 2): nan,
        (10, 3): nan,
        (10, 4): nan,
        (10, 5): nan,
        (10, 6): 13,
        (10, 7): 14,
        (10, 8): 15,
        (10, 9): 16,
        (10, 10): 16,
        (10, 11): 17,
        (10, 12): 17,
        (10, 13): 18,
        (10, 14): 18,
        (10, 15): 18,
        (10, 16): 19,
        (10, 17): 19,
        (10, 18): 19,
        (10, 19): 20,
        (10, 20): 20,
        (11, 1): nan,
        (11, 2): nan,
        (11, 3): nan,
        (11, 4): nan,
        (11, 5): nan,
        (11, 6): 13,
        (11, 7): 14,
        (11, 8): 15,
        (11, 9): 16,
        (11, 10): 17,
        (11, 11): 17,
        (11, 12): 18,
        (11, 13): 19,
        (11, 14): 19,
        (11, 15): 19,
        (11, 16): 20,
        (11, 17): 20,
        (11, 18): 20,
        (11, 19): 21,
        (11, 20): 21,
        (12, 1): nan,
        (12, 2): nan,
        (12, 3): nan,
        (12, 4): nan,
        (12, 5): nan,
        (12, 6): 13,
        (12, 7): 14,
        (12, 8): 16,
        (12, 9): 16,
        (12, 10): 17,
        (12, 11): 18,
        (12, 12): 19,
        (12, 13): 19,
        (12, 14): 20,
        (12, 15): 20,
        (12, 16): 21,
        (12, 17): 21,
        (12, 18): 21,
        (12, 19): 22,
        (12, 20): 22,
        (13, 1): nan,
        (13, 2): nan,
        (13, 3): nan,
        (13, 4): nan,
        (13, 5): nan,
        (13, 6): nan,
        (13, 7): 15,
        (13, 8): 16,
        (13, 9): 17,
        (13, 10): 18,
        (13, 11): 19,
        (13, 12): 19,
        (13, 13): 20,
        (13, 14): 20,
        (13, 15): 21,
        (13, 16): 21,
        (13, 17): 22,
        (13, 18): 22,
        (13, 19): 23,
        (13, 20): 23,
        (14, 1): nan,
        (14, 2): nan,
        (14, 3): nan,
        (14, 4): nan,
        (14, 5): nan,
        (14, 6): nan,
        (14, 7): 15,
        (14, 8): 16,
        (14, 9): 17,
        (14, 10): 18,
        (14, 11): 19,
        (14, 12): 20,
        (14, 13): 20,
        (14, 14): 21,
        (14, 15): 22,
        (14, 16): 22,
        (14, 17): 23,
        (14, 18): 23,
        (14, 19): 23,
        (14, 20): 24,
        (15, 1): nan,
        (15, 2): nan,
        (15, 3): nan,
        (15, 4): nan,
        (15, 5): nan,
        (15, 6): nan,
        (15, 7): 15,
        (15, 8): 16,
        (15, 9): 18,
        (15, 10): 18,
        (15, 11): 19,
        (15, 12): 20,
        (15, 13): 21,
        (15, 14): 22,
        (15, 15): 22,
        (15, 16): 23,
        (15, 17): 23,
        (15, 18): 24,
        (15, 19): 24,
        (15, 20): 25,
        (16, 1): nan,
        (16, 2): nan,
        (16, 3): nan,
        (16, 4): nan,
        (16, 5): nan,
        (16, 6): nan,
        (16, 7): nan,
        (16, 8): 17,
        (16, 9): 18,
        (16, 10): 19,
        (16, 11): 20,
        (16, 12): 21,
        (16, 13): 21,
        (16, 14): 22,
        (16, 15): 23,
        (16, 16): 23,
        (16, 17): 24,
        (16, 18): 25,
        (16, 19): 25,
        (16, 20): 25,
        (17, 1): nan,
        (17, 2): nan,
        (17, 3): nan,
        (17, 4): nan,
        (17, 5): nan,
        (17, 6): nan,
        (17, 7): nan,
        (17, 8): 17,
        (17, 9): 18,
        (17, 10): 19,
        (17, 11): 20,
        (17, 12): 21,
        (17, 13): 22,
        (17, 14): 23,
        (17, 15): 23,
        (17, 16): 24,
        (17, 17): 25,
        (17, 18): 25,
        (17, 19): 26,
        (17, 20): 26,
        (18, 1): nan,
        (18, 2): nan,
        (18, 3): nan,
        (18, 4): nan,
        (18, 5): nan,
        (18, 6): nan,
        (18, 7): nan,
        (18, 8): 17,
        (18, 9): 18,
        (18, 10): 19,
        (18, 11): 20,
        (18, 12): 21,
        (18, 13): 22,
        (18, 14): 23,
        (18, 15): 24,
        (18, 16): 25,
        (18, 17): 25,
        (18, 18): 26,
        (18, 19): 26,
        (18, 20): 27,
        (19, 1): nan,
        (19, 2): nan,
        (19, 3): nan,
        (19, 4): nan,
        (19, 5): nan,
        (19, 6): nan,
        (19, 7): nan,
        (19, 8): 17,
        (19, 9): 18,
        (19, 10): 20,
        (19, 11): 21,
        (19, 12): 22,
        (19, 13): 23,
        (19, 14): 23,
        (19, 15): 24,
        (19, 16): 25,
        (19, 17): 26,
        (19, 18): 26,
        (19, 19): 27,
        (19, 20): 27,
        (20, 1): nan,
        (20, 2): nan,
        (20, 3): nan,
        (20, 4): nan,
        (20, 5): nan,
        (20, 6): nan,
        (20, 7): nan,
        (20, 8): 17,
        (20, 9): 18,
        (20, 10): 20,
        (20, 11): 21,
        (20, 12): 22,
        (20, 13): 23,
        (20, 14): 24,
        (20, 15): 25,
        (20, 16): 25,
        (20, 17): 26,
        (20, 18): 27,
        (20, 19): 27,
        (20, 20): 28
    }
}


_u_critical_value_table = {
    0.10: {
        (1, 9): 0,
        (1, 10): 0,
        (1, 11): 0,
        (1, 12): 0,
        (1, 13): 0,
        (1, 14): 0,
        (1, 15): 0,
        (1, 16): 0,
        (1, 17): 0,
        (1, 18): 0,
        (1, 19): 1,
        (1, 20): 1,
        (2, 3): 0,
        (2, 4): 0,
        (2, 5): 1,
        (2, 6): 1,
        (2, 7): 1,
        (2, 8): 2,
        (2, 9): 2,
        (2, 10): 3,
        (2, 11): 3,
        (2, 12): 4,
        (2, 13): 4,
        (2, 14): 5,
        (2, 15): 5,
        (2, 16): 5,
        (2, 17): 6,
        (2, 18): 6,
        (2, 19): 7,
        (2, 20): 7,
        (3, 3): 1,
        (3, 4): 1,
        (3, 5): 2,
        (3, 6): 3,
        (3, 7): 4,
        (3, 8): 5,
        (3, 9): 5,
        (3, 10): 6,
        (3, 11): 7,
        (3, 12): 8,
        (3, 13): 9,
        (3, 14): 10,
        (3, 15): 10,
        (3, 16): 11,
        (3, 17): 12,
        (3, 18): 13,
        (3, 19): 14,
        (3, 20): 15,
        (4, 4): 3,
        (4, 5): 4,
        (4, 6): 5,
        (4, 7): 6,
        (4, 8): 7,
        (4, 9): 9,
        (4, 10): 10,
        (4, 11): 11,
        (4, 12): 12,
        (4, 13): 13,
        (4, 14): 15,
        (4, 15): 16,
        (4, 16): 17,
        (4, 17): 18,
        (4, 18): 20,
        (4, 19): 21,
        (4, 20): 22,
        (5, 5): 5,
        (5, 6): 7,
        (5, 7): 8,
        (5, 8): 10,
        (5, 9): 12,
        (5, 10): 13,
        (5, 11): 15,
        (5, 12): 17,
        (5, 13): 18,
        (5, 14): 20,
        (5, 15): 22,
        (5, 16): 23,
        (5, 17): 25,
        (5, 18): 27,
        (5, 19): 28,
        (5, 20): 30,
        (6, 6): 9,
        (6, 7): 11,
        (6, 8): 13,
        (6, 9): 15,
        (6, 10): 17,
        (6, 11): 19,
        (6, 12): 21,
        (6, 13): 23,
        (6, 14): 25,
        (6, 15): 27,
        (6, 16): 29,
        (6, 17): 31,
        (6, 18): 34,
        (6, 19): 36,
        (6, 20): 38,
        (7, 7): 13,
        (7, 8): 16,
        (7, 9): 18,
        (7, 10): 21,
        (7, 11): 23,
        (7, 12): 26,
        (7, 13): 28,
        (7, 14): 31,
        (7, 15): 33,
        (7, 16): 36,
        (7, 17): 38,
        (7, 18): 41,
        (7, 19): 43,
        (7, 20): 46,
        (8, 8): 19,
        (8, 9): 22,
        (8, 10): 24,
        (8, 11): 27,
        (8, 12): 30,
        (8, 13): 33,
        (8, 14): 36,
        (8, 15): 39,
        (8, 16): 42,
        (8, 17): 45,
        (8, 18): 48,
        (8, 19): 51,
        (8, 20): 54,
        (9, 9): 25,
        (9, 10): 28,
        (9, 11): 31,
        (9, 12): 35,
        (9, 13): 38,
        (9, 14): 41,
        (9, 15): 45,
        (9, 16): 48,
        (9, 17): 52,
        (9, 18): 55,
        (9, 19): 58,
        (9, 20): 62,
        (10, 10): 32,
        (10, 11): 36,
        (10, 12): 39,
        (10, 13): 43,
        (10, 14): 47,
        (10, 15): 51,
        (10, 16): 54,
        (10, 17): 58,
        (10, 18): 62,
        (10, 19): 66,
        (10, 20): 70,
        (11, 11): 40,
        (11, 12): 44,
        (11, 13): 48,
        (11, 14): 52,
        (11, 15): 57,
        (11, 16): 61,
        (11, 17): 65,
        (11, 18): 69,
        (11, 19): 73,
        (11, 20): 78,
        (12, 12): 49,
        (12, 13): 53,
        (12, 14): 58,
        (12, 15): 63,
        (12, 16): 67,
        (12, 17): 72,
        (12, 18): 77,
        (12, 19): 81,
        (12, 20): 86,
        (13, 13): 58,
        (13, 14): 63,
        (13, 15): 68,
        (13, 16): 74,
        (13, 17): 79,
        (13, 18): 84,
        (13, 19): 89,
        (13, 20): 94,
        (14, 14): 69,
        (14, 15): 74,
        (14, 16): 80,
        (14, 17): 85,
        (14, 18): 91,
        (14, 19): 97,
        (14, 20): 102,
        (15, 15): 80,
        (15, 16): 86,
        (15, 17): 92,
        (15, 18): 98,
        (15, 19): 104,
        (15, 20): 110,
        (16, 16): 93,
        (16, 17): 99,
        (16, 18): 106,
        (16, 19): 112,
        (16, 20): 119,
        (17, 17): 106,
        (17, 18): 113,
        (17, 19): 120,
        (17, 20): 127,
        (18, 18): 120,
        (18, 19): 128,
        (18, 20): 135,
        (19, 19): 145,
        (19, 20): 143,
        (20, 20): 151
    },
    0.05: {
        (1, 19): 0,
        (1, 20): 0,
        (2, 5): 0,
        (2, 6): 0,
        (2, 7): 0,
        (2, 8): 1,
        (2, 9): 1,
        (2, 10): 1,
        (2, 11): 1,
        (2, 12): 2,
        (2, 13): 2,
        (2, 14): 3,
        (2, 15): 3,
        (2, 16): 3,
        (2, 17): 3,
        (2, 18): 4,
        (2, 19): 4,
        (2, 20): 4,
        (3, 3): 0,
        (3, 4): 0,
        (3, 5): 1,
        (3, 6): 2,
        (3, 7): 2,
        (3, 8): 3,
        (3, 9): 4,
        (3, 10): 4,
        (3, 11): 5,
        (3, 12): 5,
        (3, 13): 6,
        (3, 14): 7,
        (3, 15): 7,
        (3, 16): 8,
        (3, 17): 9,
        (3, 18): 9,
        (3, 19): 10,
        (3, 20): 11,
        (4, 4): 1,
        (4, 5): 2,
        (4, 6): 3,
        (4, 7): 4,
        (4, 8): 5,
        (4, 9): 6,
        (4, 10): 7,
        (4, 11): 8,
        (4, 12): 9,
        (4, 13): 10,
        (4, 14): 11,
        (4, 15): 12,
        (4, 16): 14,
        (4, 17): 15,
        (4, 18): 16,
        (4, 19): 17,
        (4, 20): 18,
        (5, 5): 4,
        (5, 6): 5,
        (5, 7): 6,
        (5, 8): 8,
        (5, 9): 9,
        (5, 10): 11,
        (5, 11): 12,
        (5, 12): 13,
        (5, 13): 15,
        (5, 14): 16,
        (5, 15): 18,
        (5, 16): 19,
        (5, 17): 20,
        (5, 18): 22,
        (5, 19): 23,
        (5, 20): 25,
        (6, 6): 7,
        (6, 7): 8,
        (6, 8): 10,
        (6, 9): 12,
        (6, 10): 11,
        (6, 11): 12,
        (6, 12): 13,
        (6, 13): 15,
        (6, 14): 16,
        (6, 15): 18,
        (6, 16): 19,
        (6, 17): 20,
        (6, 18): 22,
        (6, 19): 23,
        (6, 20): 25,
        (7, 7): 11,
        (7, 8): 13,
        (7, 9): 15,
        (7, 10): 17,
        (7, 11): 19,
        (7, 12): 21,
        (7, 13): 24,
        (7, 14): 26,
        (7, 15): 28,
        (7, 16): 30,
        (7, 17): 33,
        (7, 18): 35,
        (7, 19): 37,
        (7, 20): 39,
        (8, 8): 15,
        (8, 9): 18,
        (8, 10): 20,
        (8, 11): 23,
        (8, 12): 26,
        (8, 13): 28,
        (8, 14): 31,
        (8, 15): 33,
        (8, 16): 36,
        (8, 17): 39,
        (8, 18): 41,
        (8, 19): 44,
        (8, 20): 47,
        (9, 9): 21,
        (9, 10): 24,
        (9, 11): 27,
        (9, 12): 30,
        (9, 13): 33,
        (9, 14): 36,
        (9, 15): 39,
        (9, 16): 42,
        (9, 17): 45,
        (9, 18): 48,
        (9, 19): 51,
        (9, 20): 54,
        (10, 10): 27,
        (10, 11): 31,
        (10, 12): 34,
        (10, 13): 37,
        (10, 14): 41,
        (10, 15): 44,
        (10, 16): 48,
        (10, 17): 51,
        (10, 18): 55,
        (10, 19): 58,
        (10, 20): 62,
        (11, 11): 34,
        (11, 12): 38,
        (11, 13): 42,
        (11, 14): 46,
        (11, 15): 50,
        (11, 16): 54,
        (11, 17): 57,
        (11, 18): 61,
        (11, 19): 65,
        (11, 20): 69,
        (12, 12): 42,
        (12, 13): 47,
        (12, 14): 51,
        (12, 15): 55,
        (12, 16): 60,
        (12, 17): 64,
        (12, 18): 68,
        (12, 19): 72,
        (12, 20): 77,
        (13, 13): 51,
        (13, 14): 56,
        (13, 15): 61,
        (13, 16): 65,
        (13, 17): 70,
        (13, 18): 75,
        (13, 19): 80,
        (13, 20): 84,
        (14, 14): 61,
        (14, 15): 66,
        (14, 16): 71,
        (14, 17): 77,
        (14, 18): 82,
        (14, 19): 87,
        (14, 20): 92,
        (15, 15): 72,
        (15, 16): 77,
        (15, 17): 83,
        (15, 18): 88,
        (15, 19): 94,
        (15, 20): 100,
        (16, 16): 83,
        (16, 17): 89,
        (16, 18): 95,
        (16, 19): 101,
        (16, 20): 107,
        (17, 17): 96,
        (17, 18): 102,
        (17, 19): 109,
        (17, 20): 115,
        (18, 18): 109,
        (18, 19): 116,
        (18, 20): 123,
        (19, 19): 123,
        (19, 20): 130,
        (20, 20): 138
    },
    0.025: {
        (2, 8): 0,
        (2, 9): 0,
        (2, 10): 0,
        (2, 11): 0,
        (2, 12): 1,
        (2, 13): 1,
        (2, 14): 1,
        (2, 15): 1,
        (2, 16): 1,
        (2, 17): 2,
        (2, 18): 2,
        (2, 19): 2,
        (2, 20): 2,
        (3, 5): 0,
        (3, 6): 1,
        (3, 7): 1,
        (3, 8): 2,
        (3, 9): 2,
        (3, 10): 3,
        (3, 11): 3,
        (3, 12): 4,
        (3, 13): 4,
        (3, 14): 5,
        (3, 15): 5,
        (3, 16): 6,
        (3, 17): 6,
        (3, 18): 7,
        (3, 19): 7,
        (3, 20): 8,
        (4, 4): 0,
        (4, 5): 1,
        (4, 6): 2,
        (4, 7): 3,
        (4, 8): 4,
        (4, 9): 4,
        (4, 10): 5,
        (4, 11): 6,
        (4, 12): 7,
        (4, 13): 8,
        (4, 14): 9,
        (4, 15): 10,
        (4, 16): 11,
        (4, 17): 11,
        (4, 18): 12,
        (4, 19): 13,
        (4, 20): 14,
        (5, 5): 2,
        (5, 6): 3,
        (5, 7): 5,
        (5, 8): 6,
        (5, 9): 7,
        (5, 10): 8,
        (5, 11): 9,
        (5, 12): 11,
        (5, 13): 12,
        (5, 14): 13,
        (5, 15): 14,
        (5, 16): 15,
        (5, 17): 17,
        (5, 18): 19,
        (5, 20): 20,
        (6, 6): 5,
        (6, 7): 6,
        (6, 8): 8,
        (6, 9): 10,
        (6, 10): 11,
        (6, 11): 13,
        (6, 12): 14,
        (6, 13): 16,
        (6, 14): 17,
        (6, 15): 19,
        (6, 16): 21,
        (6, 17): 22,
        (6, 18): 24,
        (6, 19): 25,
        (6, 20): 27,
        (7, 7): 8,
        (7, 8): 10,
        (7, 9): 12,
        (7, 10): 14,
        (7, 11): 16,
        (7, 12): 18,
        (7, 13): 20,
        (7, 14): 22,
        (7, 15): 24,
        (7, 16): 26,
        (7, 17): 28,
        (7, 18): 30,
        (7, 19): 32,
        (7, 20): 34,
        (8, 8): 13,
        (8, 9): 15,
        (8, 10): 17,
        (8, 11): 19,
        (8, 12): 22,
        (8, 13): 24,
        (8, 14): 26,
        (8, 15): 29,
        (8, 16): 31,
        (8, 17): 34,
        (8, 18): 36,
        (8, 19): 38,
        (8, 20): 41,
        (9, 9): 17,
        (9, 10): 20,
        (9, 11): 23,
        (9, 12): 26,
        (9, 13): 28,
        (9, 14): 31,
        (9, 15): 34,
        (9, 16): 37,
        (9, 17): 39,
        (9, 18): 42,
        (9, 19): 45,
        (9, 20): 48,
        (10, 10): 23,
        (10, 11): 26,
        (10, 12): 29,
        (10, 13): 33,
        (10, 14): 36,
        (10, 15): 39,
        (10, 16): 42,
        (10, 17): 45,
        (10, 18): 48,
        (10, 19): 52,
        (10, 20): 55,
        (11, 11): 30,
        (11, 12): 33,
        (11, 13): 37,
        (11, 14): 40,
        (11, 15): 44,
        (11, 16): 47,
        (11, 17): 51,
        (11, 18): 55,
        (11, 19): 58,
        (11, 20): 62,
        (12, 12): 37,
        (12, 13): 41,
        (12, 14): 45,
        (12, 15): 49,
        (12, 16): 53,
        (12, 17): 57,
        (12, 18): 61,
        (12, 19): 65,
        (12, 20): 69,
        (13, 13): 45,
        (13, 14): 50,
        (13, 15): 54,
        (13, 16): 59,
        (13, 17): 63,
        (13, 18): 67,
        (13, 19): 72,
        (13, 20): 76,
        (14, 14): 55,
        (14, 15): 59,
        (14, 16): 64,
        (14, 17): 69,
        (14, 18): 74,
        (14, 19): 78,
        (14, 20): 83,
        (15, 15): 64,
        (15, 16): 70,
        (15, 17): 75,
        (15, 18): 80,
        (15, 19): 85,
        (15, 20): 90,
        (16, 16): 75,
        (16, 17): 81,
        (16, 18): 86,
        (16, 19): 92,
        (16, 20): 98,
        (17, 17): 87,
        (17, 18): 93,
        (17, 19): 99,
        (17, 20): 105,
        (18, 18): 99,
        (18, 19): 106,
        (18, 20): 112,
        (19, 19): 113,
        (19, 20): 119,
        (20, 20): 127
    },
    0.01: {
        (2, 13): 0,
        (2, 14): 0,
        (2, 15): 0,
        (2, 16): 0,
        (2, 17): 0,
        (2, 18): 0,
        (2, 19): 1,
        (2, 20): 1,
        (3, 7): 0,
        (3, 8): 0,
        (3, 9): 1,
        (3, 10): 1,
        (3, 11): 1,
        (3, 12): 2,
        (3, 13): 2,
        (3, 14): 2,
        (3, 15): 3,
        (3, 16): 3,
        (3, 17): 4,
        (3, 18): 4,
        (3, 19): 4,
        (3, 20): 5,
        (4, 5): 0,
        (4, 6): 1,
        (4, 7): 1,
        (4, 8): 2,
        (4, 9): 3,
        (4, 10): 3,
        (4, 11): 4,
        (4, 12): 5,
        (4, 13): 5,
        (4, 14): 6,
        (4, 15): 7,
        (4, 16): 7,
        (4, 17): 8,
        (4, 18): 9,
        (4, 19): 9,
        (4, 20): 10,
        (5, 5): 1,
        (5, 6): 2,
        (5, 7): 3,
        (5, 8): 4,
        (5, 9): 5,
        (5, 10): 6,
        (5, 11): 7,
        (5, 12): 8,
        (5, 13): 9,
        (5, 14): 10,
        (5, 15): 11,
        (5, 16): 12,
        (5, 17): 13,
        (5, 18): 14,
        (5, 19): 15,
        (5, 20): 16,
        (6, 6): 3,
        (6, 7): 4,
        (6, 8): 6,
        (6, 9): 7,
        (6, 10): 8,
        (6, 11): 9,
        (6, 12): 11,
        (6, 13): 12,
        (6, 14): 13,
        (6, 15): 15,
        (6, 16): 16,
        (6, 17): 18,
        (6, 18): 19,
        (6, 19): 20,
        (6, 20): 22,
        (7, 7): 6,
        (7, 8): 7,
        (7, 9): 9,
        (7, 10): 11,
        (7, 11): 12,
        (7, 12): 14,
        (7, 13): 16,
        (7, 14): 17,
        (7, 15): 19,
        (7, 16): 21,
        (7, 17): 23,
        (7, 18): 24,
        (7, 19): 26,
        (7, 20): 28,
        (8, 8): 9,
        (8, 9): 11,
        (8, 10): 13,
        (8, 11): 15,
        (8, 12): 17,
        (8, 13): 20,
        (8, 14): 22,
        (8, 15): 24,
        (8, 16): 26,
        (8, 17): 28,
        (8, 18): 30,
        (8, 19): 32,
        (8, 20): 34,
        (9, 9): 14,
        (9, 10): 16,
        (9, 11): 18,
        (9, 12): 21,
        (9, 13): 23,
        (9, 14): 26,
        (9, 15): 28,
        (9, 16): 31,
        (9, 17): 33,
        (9, 18): 36,
        (9, 19): 38,
        (9, 20): 40,
        (10, 10): 19,
        (10, 11): 22,
        (10, 12): 24,
        (10, 13): 27,
        (10, 14): 30,
        (10, 15): 33,
        (10, 16): 36,
        (10, 17): 38,
        (10, 18): 41,
        (10, 19): 44,
        (10, 20): 47,
        (11, 11): 25,
        (11, 12): 28,
        (11, 13): 31,
        (11, 14): 34,
        (11, 15): 37,
        (11, 16): 41,
        (11, 17): 44,
        (11, 18): 47,
        (11, 19): 50,
        (11, 20): 53,
        (12, 12): 31,
        (12, 13): 35,
        (12, 14): 38,
        (12, 15): 42,
        (12, 16): 46,
        (12, 17): 49,
        (12, 18): 53,
        (12, 19): 56,
        (12, 20): 60,
        (13, 13): 39,
        (13, 14): 43,
        (13, 15): 47,
        (13, 16): 51,
        (13, 17): 55,
        (13, 18): 59,
        (13, 19): 53,
        (13, 20): 60,
        (14, 14): 47,
        (14, 15): 51,
        (14, 16): 56,
        (14, 17): 60,
        (14, 18): 65,
        (14, 19): 69,
        (14, 20): 73,
        (15, 15): 56,
        (15, 16): 61,
        (15, 17): 66,
        (15, 18): 70,
        (15, 19): 75,
        (15, 20): 80,
        (16, 16): 66,
        (16, 17): 71,
        (16, 18): 76,
        (16, 19): 82,
        (16, 20): 87,
        (17, 17): 77,
        (17, 18): 82,
        (17, 19): 88,
        (17, 20): 93,
        (18, 18): 88,
        (18, 19): 94,
        (18, 20): 100,
        (19, 19): 101,
        (19, 20): 107,
        (20, 20): 114
    }
}


_w_critical_value_table = {
    'one-tail': {
        0.01: {
            5: nan,
            6: nan,
            7: 0,
            8: 1,
            9: 3,
            10: 5,
            11: 7,
            12: 9,
            13: 12,
            14: 15,
            15: 19,
            16: 23,
            17: 27,
            18: 32,
            19: 37,
            20: 43,
            21: 49,
            22: 55,
            23: 62,
            24: 69,
            25: 76,
            26: 84,
            27: 92,
            28: 101,
            29: 110,
            30: 120
        },
        0.05: {
            5: 0,
            6: 2,
            7: 3,
            8: 5,
            9: 8,
            10: 10,
            11: 13,
            12: 17,
            13: 21,
            14: 25,
            15: 30,
            16: 35,
            17: 41,
            18: 47,
            19: 53,
            20: 60,
            21: 67,
            22: 75,
            23: 83,
            24: 91,
            25: 100,
            26: 110,
            27: 119,
            28: 130,
            29: 140,
            30: 151
        },
        0.025: {
            5: nan,
            6: 0,
            7: 2,
            8: 3,
            9: 5,
            10: 8,
            11: 10,
            12: 13,
            13: 17,
            14: 21,
            15: 25,
            16: 29,
            17: 34,
            18: 40,
            19: 46,
            20: 52,
            21: 58,
            22: 65,
            23: 73,
            24: 81,
            25: 89,
            26: 98,
            27: 107,
            28: 116,
            29: 126,
            30: 137
        },
        0.005: {
            5: nan,
            6: nan,
            7: nan,
            8: 0,
            9: 1,
            10: 3,
            11: 5,
            12: 7,
            13: 9,
            14: 12,
            15: 15,
            16: 19,
            17: 23,
            18: 27,
            19: 32,
            20: 37,
            21: 42,
            22: 48,
            23: 54,
            24: 61,
            25: 68,
            26: 75,
            27: 83,
            28: 91,
            29: 100,
            30: 109
        }
    },
    'two-tail': {
        0.01: {
            5: nan,
            6: nan,
            7: nan,
            8: 0,
            9: 1,
            10: 3,
            11: 5,
            12: 7,
            13: 9,
            14: 12,
            15: 15,
            16: 19,
            17: 23,
            18: 27,
            19: 32,
            20: 37,
            21: 42,
            22: 48,
            23: 54,
            24: 61,
            25: 68,
            26: 75,
            27: 83,
            28: 91,
            29: 100,
            30: 109
        },
        0.02: {
            5: nan,
            6: nan,
            7: 0,
            8: 1,
            9: 3,
            10: 5,
            11: 7,
            12: 9,
            13: 12,
            14: 15,
            15: 19,
            16: 23,
            17: 27,
            18: 32,
            19: 37,
            20: 43,
            21: 49,
            22: 55,
            23: 62,
            24: 69,
            25: 76,
            26: 84,
            27: 92,
            28: 101,
            29: 110,
            30: 120
        },
        0.05: {
            5: nan,
            6: 0,
            7: 2,
            8: 3,
            9: 5,
            10: 8,
            11: 10,
            12: 13,
            13: 17,
            14: 21,
            15: 25,
            16: 29,
            17: 34,
            18: 40,
            19: 46,
            20: 52,
            21: 58,
            22: 65,
            23: 73,
            24: 81,
            25: 89,
            26: 98,
            27: 107,
            28: 116,
            29: 126,
            30: 137
        },
        0.10: {
            5: 0,
            6: 2,
            7: 3,
            8: 5,
            9: 8,
            10: 10,
            11: 13,
            12: 17,
            13: 21,
            14: 25,
            15: 30,
            16: 35,
            17: 41,
            18: 47,
            19: 53,
            20: 60,
            21: 67,
            22: 75,
            23: 83,
            24: 91,
            25: 100,
            26: 110,
            27: 119,
            28: 130,
            29: 140,
            30: 151
        }
    }
}
